// SPDX-License-Identifier: BSD-3-Clause

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/sendfile.h>
#include <sys/eventfd.h>
#include <libaio.h>
#include <errno.h>
#include <netinet/tcp.h>

#include "aws.h"
#include "utils/util.h"
#include "utils/debug.h"
#include "utils/sock_util.h"
#include "utils/w_epoll.h"

http_parser request_parser;

char filename[BUFSIZ];

static int listenfd;

static int epollfd;

//static io_context_t ctx;

static int aws_on_path_cb(http_parser *p, const char *buf, size_t len)
{
	struct connection *conn = (struct connection *)p->data;

	memcpy(conn->request_path, buf, len);
	conn->request_path[len] = '\0';
	conn->have_path = 1;

	assert(p == &request_parser);
	memcpy(filename, buf, len);
	return 0;
}

int parse_header(struct connection **conn)
{
	/* TODO: Parse the HTTP header and extract the file path. */
	/* Use mostly null settings except for on_path callback. */
	http_parser_settings settings_on_path = {
		.on_message_begin = 0,
		.on_header_field = 0,
		.on_header_value = 0,
		.on_path = aws_on_path_cb,
		.on_url = 0,
		.on_fragment = 0,
		.on_query_string = 0,
		.on_body = 0,
		.on_headers_complete = 0,
		.on_message_complete = 0
	};

	(*conn)->parser_settings = settings_on_path;

	if (*conn == NULL)
		return -1;

	return 0;
}

struct connection *connection_create(int sockfd)
{
	/* TODO: Initialize connection structure on given socket. */
	struct connection *conn = malloc(sizeof(*conn));

	if (!conn) {
		perror("malloc");
		return NULL;
	}
	conn->sockfd = sockfd;

	memset(conn->recv_buffer, 0, BUFSIZ);
	conn->recv_len = 0;
	memset(conn->send_buffer, 0, BUFSIZ);

	conn->header_is_written = 0;

	return conn;
}

void connection_remove(struct connection *conn)
{
	/* TODO: Remove connection handler. */
	shutdown(conn->sockfd, SHUT_RDWR);

	conn->state = STATE_CONNECTION_CLOSED;
	free(conn);
}

void set_socket_options(int sockfd)
{
	int yes = 1;

	setsockopt(sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&yes, sizeof(int));

	fcntl(sockfd, F_SETFL, fcntl(sockfd, F_GETFL, 0) | O_NONBLOCK);
}

struct connection *accept_new_connection(void)
{
	struct sockaddr_in addr;

	socklen_t addrlen = sizeof(struct sockaddr_in);

	int sockfd = accept(listenfd, (struct sockaddr *)&addr, &addrlen);

	DIE(sockfd < 0, "accept");

	set_socket_options(sockfd);

	return connection_create(sockfd);
}

void add_connection_to_epoll(struct connection *conn)
{
	int rc = w_epoll_add_ptr_in(epollfd, conn->sockfd, conn);

	DIE(rc < 0, "w_epoll_add_in");
}

void handle_new_connection(void)
{
	struct connection *conn = accept_new_connection();

	add_connection_to_epoll(conn);
}

void receive_data(struct connection *conn)
{
	/* TODO: Receive message on socket.
	 * Store message in recv_buffer in struct connection.
	 */

	ssize_t bytes_recv;
	char abuffer[64];

	int rc = get_peer_address(conn->sockfd, abuffer, 64);

	if (rc < 0) {
		ERR("get_peer_address");
		connection_remove(conn);
		return;
	}

	bytes_recv = recv(conn->sockfd, conn->recv_buffer, BUFSIZ, 0);
	if (bytes_recv < 0) {
		connection_remove(conn);
		return;
	}
	if (bytes_recv == 0) {
		connection_remove(conn);
		return;
	}

	conn->recv_len = bytes_recv;
	conn->state = STATE_DATA_RECEIVED;
}

enum connection_state connection_send_static(struct connection *conn)
{
	/* TODO: Send static data using sendfile(2). */
	if (conn == NULL)
		return STATE_CONNECTION_CLOSED;

	off_t sent_bytes = 0;

	while (sent_bytes < conn->file_size) {
		ssize_t rc = sendfile(conn->sockfd, conn->fd, NULL,
							  conn->file_size - sent_bytes);

		if (rc < 0) {
			perror("Error in sendfile");
			return STATE_CONNECTION_CLOSED;
		}
		sent_bytes += rc;
	}

	conn->state = STATE_DATA_SENT;
	return STATE_DATA_SENT;
}

void free_io_array(struct iocb **iocb_array, int size)
{
	for (int i = 0; i < size; i++)
		free(iocb_array[i]);

	free(iocb_array);
}

void free_data_blocks(char **data_blocks, int size)
{
	for (int i = 0; i < size; i++)
		free(data_blocks[i]);

	free(data_blocks);
}

void free_the_io(struct connection *conn)
{
	int readb_sz = MIN(conn->file_size, BUFSIZ);

	free_io_array(conn->iocb_r, conn->file_size / readb_sz);
	free_io_array(conn->iocb_w, conn->file_size / readb_sz);
	free_data_blocks(conn->data_blocks, conn->file_size / readb_sz);

	free(conn->events);

	io_destroy(conn->ctx);
}

void set_iocb(struct connection *conn)
{
	conn->eventfd = eventfd(0, 0);
	if (conn->eventfd < 0) {
		perror("Invalid eventfd!\n");
		return;
	}

	int readb_sz = MIN(conn->file_size, BUFSIZ);

	conn->iocb_r = malloc(readb_sz * sizeof(struct iocb *));
	conn->iocb_w = malloc(readb_sz * sizeof(struct iocb *));
	conn->data_blocks = malloc(readb_sz * sizeof(char *));
	conn->events = malloc(sizeof(struct io_event));

	if (conn->iocb_r == NULL || conn->iocb_w == NULL || conn->data_blocks == NULL || conn->events == NULL) {
		perror("Error while allocating memory.\n");
		return;
	}

	for (int i = 0; i < conn->file_size / readb_sz; i++) {
		conn->iocb_r[i] = malloc(sizeof(struct iocb));
		conn->iocb_w[i] = malloc(sizeof(struct iocb));
		conn->data_blocks[i] = malloc(BUFSIZ * sizeof(char));

		if (conn->iocb_r[i] == NULL || conn->iocb_w[i] == NULL || conn->data_blocks[i] == NULL) {
			perror("Error while allocating memory.\n");
			return;
		}

		memset(conn->iocb_r[i], 0, sizeof(struct iocb));
		memset(conn->iocb_w[i], 0, sizeof(struct iocb));
	}

	conn->ctx = 0;
	if (io_setup(NUM_OPS, &conn->ctx) < 0) {
		perror("Invalid io_setup.\n");
		return;
	}
}

void connection_start_async_io(struct connection *conn)
{
	/* TODO: Start asynchronous operation (read from file).
	 * Use io_submit(2) & friends for reading data asynchronously.
	 */

	if (conn == NULL || conn->fd < 0)
		return;

	io_prep_pread(&conn->iocb, conn->fd, conn->recv_buffer, BUFSIZ, 0);
	conn->piocb[0] = &conn->iocb;

	if (io_submit(conn->ctx, 1, conn->piocb) < 0) {
		perror("io_submit failed");
		return;
	}
}

void connection_complete_async_io(struct connection *conn)
{
	/* TODO: Complete asynchronous operation; operation returns successfully.
	 * Prepare socket for sending.
	 */
	if (conn == NULL)
		return;

	int events = io_getevents(conn->ctx, 1, 1, conn->events, NULL);

	if (events < 0) {
		perror("io_getevents failed");
		return;
	}

	memcpy(conn->send_buffer, conn->recv_buffer, BUFSIZ);
	conn->send_len = BUFSIZ;
	conn->state = STATE_ASYNC_ONGOING;
}

int connection_send_dynamic(struct connection *conn)
{
	/* TODO: Read data asynchronously.
	 * Returns 0 on success and -1 on error.
	 */
	if (conn == NULL)
		return -1;

	set_iocb(conn);

	int readb_sz = MIN(conn->file_size, BUFSIZ);

	int i = 0;

	while (i < conn->file_size / readb_sz) {
		io_prep_pread(conn->iocb_r[i], conn->fd, conn->data_blocks[i], readb_sz, i * readb_sz);
		io_set_eventfd(conn->iocb_r[i], conn->eventfd);

		if (io_submit(conn->ctx, 1, &conn->iocb_r[i]) < 0) {
			free_the_io(conn);
			return -1;
		}

		connection_start_async_io(conn);
		for (int no_events_read = 0; no_events_read < NUM_OPS; no_events_read++) {
			no_events_read = io_getevents(conn->ctx, NUM_OPS, NUM_OPS, conn->events, NULL);
			DIE(no_events_read < 0, "Invalid io_getevents\n");
		}
		io_prep_pwrite(conn->iocb_w[i], conn->sockfd, conn->data_blocks[i], readb_sz, 0);

		if (io_submit(conn->ctx, 1, &conn->iocb_w[i]) < 0) {
			free_the_io(conn);
			return -1;
		}

		io_set_eventfd(conn->iocb_w[i], conn->eventfd);

		connection_complete_async_io(conn);
		memset(conn->events, 0, sizeof(struct io_event));

		i++;
	}

	free_the_io(conn);
	conn->state = STATE_DATA_SENT;

	return 0;
}

int connection_send_data(struct connection *conn)
{
	/* May be used as a helper function. */
	/* TODO: Send as much data as possible from the connection send buffer.
	 * Returns the number of bytes sent or -1 if an error occurred
	 */
	int rc;

	if (conn->file_type == STATIC) {
		rc = connection_send_static(conn);
		if (rc < 0) {
			connection_remove(conn);
			return -1;
		}
	} else if (conn->file_type == DYNAMIC) {
		int rc = connection_send_dynamic(conn);

		if (rc < 0) {
			connection_remove(conn);
			return -1;
		}
	}

	return rc;
}

void remove_connection_for_epoll(struct connection *conn)
{
	int rc = w_epoll_remove_ptr(epollfd, conn->sockfd, conn);

	if (rc < 0) {
		connection_remove(conn);
		return;
	}
}

static enum connection_state message_send(struct connection *conn)
{
	if (conn->state == STATE_CONNECTION_CLOSED)
		return STATE_CONNECTION_CLOSED;

	ssize_t total_sent_bytes = 0;

	while (!conn->header_is_written && total_sent_bytes < conn->send_len) {
		ssize_t bytes_sent = send(conn->sockfd, conn->send_buffer + total_sent_bytes,
								conn->send_len - total_sent_bytes, 0);

		if (bytes_sent == 0) {
			remove_connection_for_epoll(conn);
			return STATE_CONNECTION_CLOSED;
		}

		total_sent_bytes += bytes_sent;
	}

	conn->header_is_written = 1;

	if (conn->header_is_written && conn->fd != FILE_NOT_FOUND) {
		int rc = connection_send_data(conn);

		if (rc < 0) {
			connection_remove(conn);
			return STATE_CONNECTION_CLOSED;
		}
	}

	if (conn->state == STATE_DATA_SENT) {
		w_epoll_remove_ptr(epollfd, conn->sockfd, conn);
		remove_connection_for_epoll(conn);
		return STATE_CONNECTION_CLOSED;
	}

	w_epoll_update_ptr_in(epollfd, conn->sockfd, conn);
	conn->state = STATE_DATA_SENT;

	return STATE_DATA_SENT;
}

const char *extract_file_type_from_filename(const char *filename)
{
	const char *start = filename + 1;
	const char *end = strchr(start, '/');

	if (end == NULL)
		return NULL;

	size_t length = (size_t)(end - start);
	char *file_type_str = malloc(length + 1);

	if (file_type_str == NULL) {
		perror("Error while allocating memory for file type string.\n");
		return NULL;
	}

	memcpy(file_type_str, start, length);
	file_type_str[length] = '\0';

	return file_type_str;
}

int map_file_type_string_to_enum(const char *file_type_str)
{
	if (strcmp(file_type_str, "static") == 0)
		return STATIC;
	else if (strcmp(file_type_str, "dynamic") == 0)
		return DYNAMIC;
	else
		return FILE_NOT_FOUND;
}

int connection_open_file(struct connection *conn)
{
	/* TODO: Open file and update connection fields. */
	if (conn == NULL || conn->filename[0] == '\0')
		return -1;

	conn->fd = open(conn->filename, O_RDONLY);

	if (conn->fd < 0) {
		perror("Failed to open file");
		return -1;
	}

	return 0;
}

void prepare_connection_file_data(struct connection *conn)
{
	if (strlen(filename) == 0)
		return;

	memset(conn->filename, 0, BUFSIZ);
	memmove(conn->filename, AWS_DOCUMENT_ROOT, strlen(AWS_DOCUMENT_ROOT));
	memmove(conn->filename + strlen(AWS_DOCUMENT_ROOT), filename + 1, strlen(filename) - 1);

	if (!strstr(filename, "."))
		strcat(conn->filename, ".");

	strcat(conn->filename, "dat");

	if (connection_open_file(conn) == -1) {
		perror("Error opening file");
		conn->fd = -1;
		conn->file_type = FILE_NOT_FOUND;
	} else {
		conn->fd = open(conn->filename, O_RDWR);
		const char *file_type_str = extract_file_type_from_filename(filename);

		if (file_type_str == NULL) {
			conn->file_type = FILE_NOT_FOUND;
			return;
		}

		conn->file_type = map_file_type_string_to_enum(file_type_str);
	}
}

static void connection_prepare_send_reply_header(struct connection *conn)
{
	/* TODO: Prepare the connection buffer to send the reply header. */
	if (conn == NULL)
		return;

	const char *reply_header = HTTP_OK_MSG;

	memcpy(conn->send_buffer, reply_header, sizeof(conn->send_buffer) - 1);
	conn->send_buffer[sizeof(conn->send_buffer) - 1] = '\0';
	conn->send_len = strlen(conn->send_buffer);
}

static void connection_prepare_send_404(struct connection *conn)
{
	/* TODO: Prepare the connection buffer to send the 404 header. */
	if (conn == NULL)
		return;

	const char *not_found_header = HTTP_NOT_FOUND_MSG;

	memcpy(conn->send_buffer, not_found_header, sizeof(conn->send_buffer) - 1);
	conn->send_len = strlen(conn->send_buffer);
}

void update_connection_send_buffer(struct connection *conn, int file_not_found)
{
	memset(conn->send_buffer, 0, BUFSIZ);

	const char *status_msg = file_not_found ? HTTP_NOT_FOUND_MSG : HTTP_OK_MSG;

	conn->send_len = strlen(status_msg);
	memmove(conn->send_buffer, status_msg, conn->send_len);
	conn->send_buffer[conn->send_len] = '\0';

	if (!file_not_found) {
		off_t current_position = lseek(conn->fd, 0, SEEK_CUR);

		conn->file_size = lseek(conn->fd, 0, SEEK_END);
		lseek(conn->fd, current_position, SEEK_SET);
	} else {
		conn->file_size = 0;
	}
}

enum resource_type connection_get_resource_type(struct connection *conn)
{
	/* TODO: Get resource type depending on request path/filename. Filename should
	 * point to the static or dynamic folder.
	 */

	if (conn == NULL || conn->request_path[0] == '\0')
		return RESOURCE_TYPE_NONE;

	char *saveptr;
	char *first_part = strtok_r(conn->request_path, "/", &saveptr);

	if (first_part == NULL)
		return RESOURCE_TYPE_NONE;

	if (strcmp(first_part, AWS_REL_STATIC_FOLDER) == 0)
		return RESOURCE_TYPE_STATIC;
	else if (strcmp(first_part, AWS_REL_DYNAMIC_FOLDER) == 0)
		return RESOURCE_TYPE_DYNAMIC;
	else
		return RESOURCE_TYPE_NONE;
}

static void process_client_request(struct connection *conn)
{
	if (conn == NULL)
		return;

	receive_data(conn);
	if (conn->state != STATE_DATA_RECEIVED)
		return;

	int rc;

	http_parser_init(&request_parser, HTTP_REQUEST);
	request_parser.data = conn;
	rc = parse_header(&conn);
	if (rc < 0) {
		connection_remove(conn);
		return;
	}
	http_parser_execute(&request_parser, &conn->parser_settings, conn->recv_buffer, conn->recv_len);

	prepare_connection_file_data(conn);
	if (conn->fd == FILE_NOT_FOUND) {
		connection_prepare_send_404(conn);
	} else {
		connection_prepare_send_reply_header(conn);

		off_t s = lseek(conn->fd, 0, SEEK_CUR);

		conn->file_size = lseek(conn->fd, 0, SEEK_END);

		lseek(conn->fd, s, SEEK_SET);
	}

	rc = w_epoll_update_ptr_inout(epollfd, conn->sockfd, conn);
	if (rc < 0)
		connection_remove(conn);
}

void handle_input(struct connection *conn)
{
	/* TODO: Handle input information: may be a new message or notification of
	 * completion of an asynchronous I/O operation.
	 */
	if (conn == NULL)
		return;

	switch (conn->state) {
	case STATE_DATA_RECEIVED:
		process_client_request(conn);
		break;
	default:
		printf("shouldn't get here %d\n", conn->state);
		break;
	}
}

void handle_output(struct connection *conn)
{
	/* TODO: Handle output information: may be a new valid requests or notification of
	 * completion of an asynchronous I/O operation or invalid requests.
	 */
	if (conn == NULL)
		return;

	switch (conn->state) {
	case STATE_DATA_SENT:
		message_send(conn);
		break;
	default:
		ERR("Unexpected state\n");
		exit(1);
	}
}

void handle_client(uint32_t event, struct connection *conn)
{
	/* TODO: Handle new client. There can be input and output connections.
	 * Take care of what happened at the end of a connection.
	 */
	switch (event) {
	case EPOLLIN:
		handle_input(conn);
		break;
	case EPOLLOUT:
		handle_output(conn);
		break;
	default:
		ERR("Unexpected event\n");
		exit(1);
	}
}

int main(void)
{
	int rc;

	epollfd = w_epoll_create();

	listenfd = tcp_create_listener(AWS_LISTEN_PORT,
		DEFAULT_LISTEN_BACKLOG);
	rc = w_epoll_add_fd_in(epollfd, listenfd);
	DIE(rc < 0, "w_epoll_add_fd_in");
	while (1) {
		struct epoll_event rev;

		rc = w_epoll_wait_infinite(epollfd, &rev);

		if (rev.data.fd == listenfd) {
			if (rev.events & EPOLLIN)
				handle_new_connection();
		} else {
			if (rev.events & EPOLLIN)
				process_client_request(rev.data.ptr);
			if (rev.events & EPOLLOUT)
				message_send(rev.data.ptr);
		}
	}

	return 0;
}
